
-- Utility functions

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function ItemScore:CheckDB()
	if not self.db.char.scoreSets then
		-- create default score set
		self.db.char.scoreSets = {}
		
		self.db.char.scoreSets["Default"] = self:CreateEmptyScoreSet("Default")
	end
	
	-- default trinket clauses here
	if not self.db.global.trinketClauses then
		self:ResetTrinketClauses()
	end
	
	-- default trinket options to display
	if self.db.char.showTrinketUptime == nil then
		self.db.char.showTrinketUptime = true
	end
	
	if self.db.char.showTrinketEffect == nil then
		self.db.char.showTrinketEffect = true
	end
	
	-- migrate existing score sets
	local scoreSets = self:GetScoreSets()
	for i in pairs(scoreSets) do
		self:MigrateScoreSet(i)
	end
	
	-- migrate existing trinket clauses
	self:MigrateTrinketClauses()
end

function ItemScore:MigrateScoreSet(name)
	local scoreSet = self:GetScoreSet(name)
	
	-- initialise calcTrinket if it is not already set
	if scoreSet.calcTrinket == nil then
		scoreSet.calcTrinket = true
	end
	
	-- initialise trinket categories if not already set
	if not scoreSet.trinketCategories then
		scoreSet.trinketCategories = { ["All"] = true }
	end
end

function ItemScore:CreateEmptyScoreSet(name)
	local ret = {}
	
	ret.name = name
	ret.display = 1
	ret.r = 0.5
	ret.g = 0.5
	ret.b = 0.5 -- default to grey
	ret.pattern = "%n: %a (%d) [%p] %t" -- Default
	ret.precision = 2
	ret.calcTrinket = true
	ret.info = {}
	ret.trinketCategories = { ["All"] = true }
	
	return ret
end

function ItemScore:CreateNewSet(newName)
	local scoreSets = self:GetScoreSets()
	local newSet = self:CreateEmptyScoreSet(newName)
	scoreSets[newName] = newSet
	return newSet
end

function ItemScore:CopySet(existingSetName, newName)
	local scoreSets = self:GetScoreSets()
	local existingSet = scoreSets[existingSetName]
	
	local newSet = deepcopy(existingSet)
	newSet.name = newName 
	scoreSets[newName] = newSet
	
	return newSet
end

function ItemScore:RenameSet(existingSetName, newName)
	local scoreSets = self:GetScoreSets()
	scoreSets[newName] = scoreSets[existingSetName]
	scoreSets[existingSetName] = nil
		
	-- need to check all sets and ensure that any that use it as a parent have 
	-- the parent name changed
	for i in pairs(scoreSets) do
		if scoreSets[i].parent == existingSetName then
			scoreSets[i].parent = newName
		end
	end
	
	scoreSets[newName].name = newName
end

function ItemScore:DeleteSet(setName)
	local scoreSets = self:GetScoreSets()
	scoreSets[setName] = nil
	
	-- need to check all sets and ensure that any that use it as a parent have 
	-- the parent blanked
	for i in pairs(scoreSets) do
		if scoreSets[i].parent == setName then
			scoreSets[i].parent = nil
		end
	end
end

function ItemScore:GetScore(setName, statName)
	local scoreSet = self:GetScoreSet(setName)
	return scoreSet.info[statName].score
end

function ItemScore:SetScore(setName, statName, value)
	-- ensure if number
	if not tonumber(value) then
		DEFAULT_CHAT_FRAME:AddMessage("Cannot set stat '" .. statName .. "' to non-numeric value " .. value .. ".")
	end
	
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.info[statName].score = value
end

function ItemScore:GetCap(setName, statName)
	local scoreSet = self:GetScoreSet(setName)
	return scoreSet.info[statName].cap
end

function ItemScore:SetCap(setName, statName, value)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.info[statName].cap = value
end

function ItemScore:AddStat(setName, statName)
	local scoreSet = self:GetScoreSet(setName)
	
	-- ensure that stat is supported
	if not ItemScore.supportedStats[statName] then
		DEFAULT_CHAT_FRAME:AddMessage("Unsupported stat '" .. statName .. "' ignored.")
		return false
	end
	
	scoreSet.info[statName] = {}
	scoreSet.info[statName].score = 0
	scoreSet.info[statName].cap = -1
	
	return true
end

function ItemScore:RemoveStat(setName, statName)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.info[statName] = nil
end

function ItemScore:GetParent(setName)
	local scoreSet = self:GetScoreSet(setName)
	
	return scoreSet.parent
end

function ItemScore:SetParent(setName, parentValue)
	local scoreSet = self:GetScoreSet(setName)
	
	if parentValue and parentValue ~= "" then
	    scoreSet.parent = parentValue
	else
	    scoreSet.parent = nil
	end 
end

function ItemScore:GetPrecision(setName, value)
	local scoreSet = self:GetScoreSet(setName)
	return scoreSet.precision
end

function ItemScore:SetPrecision(setName, value)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.precision = value
end

function ItemScore:GetEnabled(setName)
	local scoreSet = self:GetScoreSet(setName)
	
	return scoreSet.display
end

function ItemScore:SetEnabled(setName, value)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.display = value
end

function ItemScore:GetCalcTrinket(setName)
	local scoreSet = self:GetScoreSet(setName)
	
	return scoreSet.calcTrinket
end

function ItemScore:SetCalcTrinket(setName, value)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.calcTrinket = value
end

function ItemScore:GetNormalise(setName)
	local scoreSet = self:GetScoreSet(setName)
	
	return scoreSet.normalise
end

function ItemScore:SetNormalise(setName, value)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.normalise = value
end

function ItemScore:TrinketCalcEnabled()
	-- iterate through all enabled score sets
	-- return true if any have trinket calc enabled
	local scoreSets = self:GetScoreSets()
	for i in pairs(scoreSets) do
		if self:GetEnabled(i) and self:GetCalcTrinket(i) then
			return true
		end
	end
	
	return false
end

function ItemScore:SetColour(setName, r, g, b, a)
	local scoreSet = self:GetScoreSet(setName)
	
	scoreSet.r = r
	scoreSet.g = g
	scoreSet.b = b
end

function ItemScore:GetColour(setName)
	local scoreSet = self:GetScoreSet(setName)
	if not scoreSet then
		-- display error here
		return 0,0,0,1
	else
		return scoreSet.r, scoreSet.g, scoreSet.b, 1.0 
	end
end

function ItemScore:GetPattern(setName)
	local scoreSet = self:GetScoreSet(setName)
		
	return scoreSet.pattern
end

function ItemScore:SetPattern(setName, value)
	local scoreSet = self:GetScoreSet(setName)
		
	scoreSet.pattern = value
end

function ItemScore:GetScoreSetCategory(setName, category)
	local scoreSet = self:GetScoreSet(setName)
	if scoreSet then
		return scoreSet.trinketCategories[category]
	end
	
	return false
end

function ItemScore:SetScoreSetCategory(setName, category, value)
	local scoreSet = self:GetScoreSet(setName)
	if scoreSet then
		scoreSet.trinketCategories[category] = value
	end
end

function ItemScore:GetScoreSets()
	return self.db.char.scoreSets
end

function ItemScore:GetScoreSet(name)
	return self:GetScoreSets()[name]
end

function ItemScore:GetDefaultSet()
	local scoreSets = self:GetScoreSets()
	for i in pairs(scoreSets) do
		return i
	end
	
	return nil
end

function ItemScore:GetEffectiveSet(name)
	local scoreSet = self:GetScoreSet(name)
	
	if not scoreSet then return nil end
	if not scoreSet.parent then return scoreSet.info end
	
	local parentSet = self:GetScoreSet(scoreSet.parent)
	local effSet = deepcopy(parentSet.info)
	
	for i,v in pairs(scoreSet.info) do
		effSet[i] = v
	end
	
	return effSet
end

function ItemScore:GetNormaliseFactor(setName)
	if not ItemScore:GetNormalise(setName) then
		return 1
	end
	
	local scoreSet = self:GetEffectiveSet(setName)
	local maxScore = 0
	for i,v in pairs(scoreSet) do
		-- don't normalise socket values
		if i ~= "EMPTY_SOCKET_RED" and i ~= "EMPTY_SOCKET_YELLOW" and
		   i ~= "EMPTY_SOCKET_BLUE" and i ~= "EMPTY_SOCKET_META" and
		   i ~= "EMPTY_SOCKET_PRISMATIC" then
			if tonumber(v.score) > maxScore then maxScore = tonumber(v.score) end
		end
	end
	return maxScore
end
